@@ -11,7 +11,7 @@
  */
 
 /* 
- * Copyright (C) 1995-2001 the Free Software Foundation, Inc.
+ * Copyright (C) 1995-2002 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -85,3 +85,7 @@
 #define HAVE_MKTIME     1
 #endif
 
+/* For whiny users */
+#ifdef USE_INCLUDED_STRFTIME
+#undef HAVE_STRFTIME
+#endif
